@@ -16,9 +16,9 @@
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * License along with this library. If not, see <http://www.gnu.org/licenses/>.
+ *
+ *
  *
  * Authors:
  *   Robert Bragg <robert@linux.intel.com>
@@ -36,6 +36,7 @@
 #include "cogl-material-private.h"
 #include "cogl-texture-private.h"
 #include "cogl-blend-string.h"
+#include "cogl-journal-private.h"
 
 #include <glib.h>
 #include <string.h>
@@ -57,6 +58,8 @@
 #define glBlendEquationSeparate ctx->drv.pf_glBlendEquationSeparate
 #endif
 
+static CoglHandle _cogl_material_layer_copy (CoglHandle layer_handle);
+
 static void _cogl_material_free (CoglMaterial *tex);
 static void _cogl_material_layer_free (CoglMaterialLayer *layer);
 
@@ -71,17 +74,19 @@ _cogl_material_error_quark (void)
   return g_quark_from_static_string ("cogl-material-error-quark");
 }
 
-CoglHandle
-cogl_material_new (void)
+void
+_cogl_material_init_default_material (void)
 {
   /* Create new - blank - material */
-  CoglMaterial *material = g_new0 (CoglMaterial, 1);
+  CoglMaterial *material = g_slice_new0 (CoglMaterial);
   GLubyte *unlit = material->unlit;
   GLfloat *ambient = material->ambient;
   GLfloat *diffuse = material->diffuse;
   GLfloat *specular = material->specular;
   GLfloat *emission = material->emission;
 
+  _COGL_GET_CONTEXT (ctx, NO_RETVAL);
+
   /* Use the same defaults as the GL spec... */
   unlit[0] = 0xff; unlit[1] = 0xff; unlit[2] = 0xff; unlit[3] = 0xff;
   material->flags |= COGL_MATERIAL_FLAG_DEFAULT_COLOR;
@@ -102,7 +107,7 @@ cogl_material_new (void)
 #ifndef HAVE_COGL_GLES
   material->blend_equation_rgb = GL_FUNC_ADD;
   material->blend_equation_alpha = GL_FUNC_ADD;
-  material->blend_src_factor_alpha = GL_SRC_ALPHA;
+  material->blend_src_factor_alpha = GL_ONE;
   material->blend_dst_factor_alpha = GL_ONE_MINUS_SRC_ALPHA;
   material->blend_constant[0] = 0;
   material->blend_constant[1] = 0;
@@ -116,9 +121,34 @@ cogl_material_new (void)
   material->layers = NULL;
   material->n_layers = 0;
 
+  ctx->default_material = _cogl_material_handle_new (material);
+}
+
+CoglHandle
+cogl_material_copy (CoglHandle handle)
+{
+  CoglMaterial *material = g_slice_new (CoglMaterial);
+  GList *l;
+
+  _COGL_GET_CONTEXT (ctx, COGL_INVALID_HANDLE);
+
+  memcpy (material, handle, sizeof (CoglMaterial));
+
+  material->layers = g_list_copy (material->layers);
+  for (l = material->layers; l; l = l->next)
+    l->data = _cogl_material_layer_copy (l->data);
+
   return _cogl_material_handle_new (material);
 }
 
+CoglHandle
+cogl_material_new (void)
+{
+  _COGL_GET_CONTEXT (ctx, COGL_INVALID_HANDLE);
+
+  return cogl_material_copy (ctx->default_material);
+}
+
 static void
 _cogl_material_free (CoglMaterial *material)
 {
@@ -128,7 +158,7 @@ _cogl_material_free (CoglMaterial *material)
   g_list_foreach (material->layers,
 		  (GFunc)cogl_handle_unref, NULL);
   g_list_free (material->layers);
-  g_free (material);
+  g_slice_free (CoglMaterial, material);
 }
 
 static gboolean
@@ -146,7 +176,11 @@ _cogl_material_needs_blending_enabled (CoglMaterial *material,
 #endif
 
   if ((override_color && override_color[3] != 0xff) ||
-      material->unlit[3] != 0xff)
+      material->unlit[3] != 0xff ||
+      material->ambient[3] != 1.0f ||
+      material->diffuse[3] != 1.0f ||
+      material->specular[3] != 1.0f ||
+      material->emission[3] != 1.0f)
     return TRUE;
 
   for (tmp = material->layers; tmp != NULL; tmp = tmp->next)
@@ -320,6 +354,8 @@ cogl_material_set_ambient (CoglHandle handle,
   ambient[3] = cogl_color_get_alpha_float (ambient_color);
 
   material->flags &= ~COGL_MATERIAL_FLAG_DEFAULT_GL_MATERIAL;
+
+  handle_automatic_blend_enable (material);
 }
 
 void
@@ -360,6 +396,8 @@ cogl_material_set_diffuse (CoglHandle handle,
   diffuse[3] = cogl_color_get_alpha_float (diffuse_color);
 
   material->flags &= ~COGL_MATERIAL_FLAG_DEFAULT_GL_MATERIAL;
+
+  handle_automatic_blend_enable (material);
 }
 
 void
@@ -408,6 +446,8 @@ cogl_material_set_specular (CoglHandle handle,
   specular[3] = cogl_color_get_alpha_float (specular_color);
 
   material->flags &= ~COGL_MATERIAL_FLAG_DEFAULT_GL_MATERIAL;
+
+  handle_automatic_blend_enable (material);
 }
 
 float
@@ -482,6 +522,8 @@ cogl_material_set_emission (CoglHandle handle,
   emission[3] = cogl_color_get_alpha_float (emission_color);
 
   material->flags &= ~COGL_MATERIAL_FLAG_DEFAULT_GL_MATERIAL;
+
+  handle_automatic_blend_enable (material);
 }
 
 void
@@ -699,7 +741,7 @@ cogl_material_set_blend_constant (CoglHandle handle,
  */
 static CoglMaterialLayer *
 _cogl_material_get_layer (CoglMaterial *material,
-			  gint          index_,
+			  int           index_,
 			  gboolean      create_if_not_found)
 {
   CoglMaterialLayer *layer;
@@ -727,7 +769,7 @@ _cogl_material_get_layer (CoglMaterial *material,
   /* possibly flush primitives referencing the current state... */
   _cogl_material_pre_change_notify (material, FALSE, NULL);
 
-  layer = g_new0 (CoglMaterialLayer, 1);
+  layer = g_slice_new0 (CoglMaterialLayer);
 
   layer_handle = _cogl_material_layer_handle_new (layer);
   layer->index = index_;
@@ -760,7 +802,7 @@ _cogl_material_get_layer (CoglMaterial *material,
 
 void
 cogl_material_set_layer (CoglHandle material_handle,
-			 gint layer_index,
+			 int layer_index,
 			 CoglHandle texture_handle)
 {
   CoglMaterial	    *material;
@@ -772,7 +814,7 @@ cogl_material_set_layer (CoglHandle material_handle,
 
   material = _cogl_material_pointer_from_handle (material_handle);
 
-  layer = _cogl_material_get_layer (material_handle, layer_index, TRUE);
+  layer = _cogl_material_get_layer (material, layer_index, TRUE);
 
   if (texture_handle == layer->texture)
     return;
@@ -781,7 +823,7 @@ cogl_material_set_layer (CoglHandle material_handle,
   _cogl_material_pre_change_notify (material, FALSE, NULL);
 
   material->n_layers = g_list_length (material->layers);
-  if (material->n_layers >= CGL_MAX_COMBINED_TEXTURE_IMAGE_UNITS)
+  if (material->n_layers >= _cogl_get_max_texture_image_units ())
     {
       if (!(material->flags & COGL_MATERIAL_FLAG_SHOWN_SAMPLER_WARNING))
 	{
@@ -893,7 +935,7 @@ setup_texture_combine_state (CoglBlendStringStatement *statement,
 
 gboolean
 cogl_material_set_layer_combine (CoglHandle handle,
-				 gint layer_index,
+				 int layer_index,
 				 const char *combine_description,
                                  GError **error)
 {
@@ -963,7 +1005,7 @@ cogl_material_set_layer_combine (CoglHandle handle,
 
 void
 cogl_material_set_layer_combine_constant (CoglHandle handle,
-				          gint layer_index,
+				          int layer_index,
                                           CoglColor *constant_color)
 {
   CoglMaterial      *material;
@@ -990,7 +1032,7 @@ cogl_material_set_layer_combine_constant (CoglHandle handle,
 
 void
 cogl_material_set_layer_matrix (CoglHandle material_handle,
-				gint layer_index,
+				int layer_index,
 				CoglMatrix *matrix)
 {
   CoglMaterial *material;
@@ -1016,12 +1058,12 @@ _cogl_material_layer_free (CoglMaterialLayer *layer)
 {
   if (layer->texture != COGL_INVALID_HANDLE)
     cogl_handle_unref (layer->texture);
-  g_free (layer);
+  g_slice_free (CoglMaterialLayer, layer);
 }
 
 void
 cogl_material_remove_layer (CoglHandle material_handle,
-			    gint layer_index)
+			    int layer_index)
 {
   CoglMaterial	     *material;
   CoglMaterialLayer  *layer;
@@ -1058,11 +1100,11 @@ cogl_material_remove_layer (CoglHandle material_handle,
 
 /* XXX: This API is hopfully just a stop-gap solution. Ideally cogl_enable
  * will be replaced. */
-gulong
+unsigned long
 _cogl_material_get_cogl_enable_flags (CoglHandle material_handle)
 {
   CoglMaterial	*material;
-  gulong	 enable_flags = 0;
+  unsigned long	 enable_flags = 0;
 
   _COGL_GET_CONTEXT (ctx, 0);
 
@@ -1128,7 +1170,7 @@ cogl_material_layer_get_texture (CoglHandle layer_handle)
   return layer->texture;
 }
 
-gulong
+unsigned long
 _cogl_material_layer_get_flags (CoglHandle layer_handle)
 {
   CoglMaterialLayer *layer;
@@ -1140,7 +1182,22 @@ _cogl_material_layer_get_flags (CoglHandle layer_handle)
   return layer->flags & COGL_MATERIAL_LAYER_FLAG_HAS_USER_MATRIX;
 }
 
-static guint
+static CoglHandle
+_cogl_material_layer_copy (CoglHandle layer_handle)
+{
+  CoglMaterialLayer *layer =
+    _cogl_material_layer_pointer_from_handle (layer_handle);
+  CoglMaterialLayer *layer_copy = g_slice_new (CoglMaterialLayer);
+
+  memcpy (layer_copy, layer, sizeof (CoglMaterialLayer));
+
+  if (layer_copy->texture != COGL_INVALID_HANDLE)
+    cogl_handle_ref (layer_copy->texture);
+
+  return _cogl_material_layer_handle_new (layer_copy);
+}
+
+static unsigned int
 get_n_args_for_combine_func (GLint func)
 {
   switch (func)
@@ -1264,6 +1321,19 @@ _cogl_material_layer_flush_gl_sampler_state (CoglMaterialLayer  *layer,
   _cogl_matrix_stack_flush_to_gl (unit->matrix_stack, COGL_MATRIX_TEXTURE);
 }
 
+void
+_cogl_material_layer_ensure_mipmaps (CoglHandle layer_handle)
+{
+  CoglMaterialLayer *layer;
+
+  layer = _cogl_material_layer_pointer_from_handle (layer_handle);
+
+  if (layer->texture &&
+      (is_mipmap_filter (layer->min_filter) ||
+       is_mipmap_filter (layer->mag_filter)))
+    _cogl_texture_ensure_mipmaps (layer->texture);
+}
+
 /*
  * _cogl_material_flush_layers_gl_state:
  * @fallback_mask: is a bitmask of the material layers that need to be
@@ -1330,11 +1400,10 @@ _cogl_material_flush_layers_gl_state (CoglMaterial *material,
       CoglHandle         tex_handle;
       GLuint             gl_texture;
       GLenum             gl_target;
-#ifdef HAVE_COGL_GLES2
-      GLenum             gl_internal_format;
-#endif
       CoglTextureUnit   *unit;
 
+      _cogl_material_layer_ensure_mipmaps (layer_handle);
+
       new_gl_layer_info.layer0_overridden =
         layer0_override_texture ? TRUE : FALSE;
       new_gl_layer_info.fallback =
@@ -1344,7 +1413,13 @@ _cogl_material_flush_layers_gl_state (CoglMaterial *material,
 
       tex_handle = layer->texture;
       if (tex_handle != COGL_INVALID_HANDLE)
-        cogl_texture_get_gl_texture (tex_handle, &gl_texture, &gl_target);
+        {
+          _cogl_texture_set_filters (tex_handle,
+                                     layer->min_filter,
+                                     layer->mag_filter);
+
+          cogl_texture_get_gl_texture (tex_handle, &gl_texture, &gl_target);
+        }
       else
         {
           new_gl_layer_info.fallback = TRUE;
@@ -1375,13 +1450,6 @@ _cogl_material_flush_layers_gl_state (CoglMaterial *material,
       GE (glActiveTexture (GL_TEXTURE0 + i));
       unit = _cogl_get_texture_unit (i);
 
-      _cogl_texture_set_filters (layer->texture,
-                                 layer->min_filter,
-                                 layer->mag_filter);
-      if (is_mipmap_filter (layer->min_filter)
-          || is_mipmap_filter (layer->mag_filter))
-        _cogl_texture_ensure_mipmaps (layer->texture);
-
       /* FIXME: We could be more clever here and only bind the texture
          if it is different from gl_layer_info->gl_texture to avoid
          redundant GL calls. However a few other places in Cogl and
@@ -1389,14 +1457,7 @@ _cogl_material_flush_layers_gl_state (CoglMaterial *material,
          we'd need to ensure they affect the cache. Also deleting a
          texture should clear it from the cache in case a new texture
          is generated with the same number */
-#ifdef HAVE_COGL_GLES2
-      gl_internal_format = _cogl_texture_get_internal_gl_format (tex_handle);
-      cogl_gles2_wrapper_bind_texture (gl_target,
-                                       gl_texture,
-                                       gl_internal_format);
-#else
       GE (glBindTexture (gl_target, gl_texture));
-#endif
 
       /* XXX: Once we add caching for glBindTexture state, these
        * checks should be moved back up to the top of the loop!
@@ -1475,7 +1536,7 @@ _cogl_material_flush_layers_gl_state (CoglMaterial *material,
 
       layer->flags &= ~COGL_MATERIAL_LAYER_FLAG_DIRTY;
 
-      if ((i+1) >= CGL_MAX_COMBINED_TEXTURE_IMAGE_UNITS)
+      if ((i+1) >= _cogl_get_max_texture_image_units ())
 	break;
     }
 
@@ -1546,16 +1607,21 @@ _cogl_material_flush_base_gl_state (CoglMaterial *material,
     {
 #if defined (HAVE_COGL_GLES2)
       gboolean have_blend_equation_seperate = TRUE;
+      gboolean have_blend_func_separate = TRUE;
 #elif defined (HAVE_COGL_GL)
       gboolean have_blend_equation_seperate = FALSE;
+      gboolean have_blend_func_separate = FALSE;
       if (ctx->drv.pf_glBlendEquationSeparate) /* Only GL 2.0 + */
         have_blend_equation_seperate = TRUE;
+      if (ctx->drv.pf_glBlendFuncSeparate) /* Only GL 1.4 + */
+        have_blend_func_separate = TRUE;
 #endif
 
 #ifndef HAVE_COGL_GLES /* GLES 1 only has glBlendFunc */
-      if (material->blend_src_factor_rgb != material->blend_src_factor_alpha
-          || (material->blend_src_factor_rgb !=
-              material->blend_src_factor_alpha))
+      if (have_blend_func_separate &&
+          (material->blend_src_factor_rgb != material->blend_src_factor_alpha ||
+           (material->blend_src_factor_rgb !=
+            material->blend_src_factor_alpha)))
         {
           if (have_blend_equation_seperate &&
               material->blend_equation_rgb != material->blend_equation_alpha)
@@ -1631,13 +1697,36 @@ _cogl_material_flush_gl_state (CoglHandle handle,
             0, sizeof (CoglMaterialFlushOptions));
 }
 
+static gboolean
+_cogl_material_texture_equal (CoglHandle texture0, CoglHandle texture1)
+{
+  GLenum gl_handle0, gl_handle1, gl_target0, gl_target1;
+
+  /* If the texture handles are the same then the textures are
+     definitely equal */
+  if (texture0 == texture1)
+    return TRUE;
+
+  /* If neither texture is sliced then they could still be the same if
+     the are referring to the same GL texture */
+  if (cogl_texture_is_sliced (texture0) ||
+      cogl_texture_is_sliced (texture1))
+    return FALSE;
+
+  cogl_texture_get_gl_texture (texture0, &gl_handle0, &gl_target0);
+  cogl_texture_get_gl_texture (texture1, &gl_handle1, &gl_target1);
+
+  return gl_handle0 == gl_handle1 && gl_target0 == gl_target1;
+}
+
 static gboolean
 _cogl_material_layer_equal (CoglMaterialLayer *material0_layer,
                             CoglHandle         material0_layer_texture,
                             CoglMaterialLayer *material1_layer,
                             CoglHandle         material1_layer_texture)
 {
-  if (material0_layer_texture != material1_layer_texture)
+  if (!_cogl_material_texture_equal (material0_layer_texture,
+                                     material1_layer_texture))
     return FALSE;
 
   if ((material0_layer->flags & COGL_MATERIAL_LAYER_FLAG_DEFAULT_COMBINE) !=
@@ -1652,6 +1741,11 @@ _cogl_material_layer_equal (CoglMaterialLayer *material0_layer,
     return FALSE;
 #endif
 
+  if (material0_layer->mag_filter != material1_layer->mag_filter)
+    return FALSE;
+  if (material0_layer->min_filter != material1_layer->min_filter)
+    return FALSE;
+
   return TRUE;
 }
 
@@ -1921,10 +2015,10 @@ cogl_set_source_texture (CoglHandle texture_handle)
 
   g_return_if_fail (texture_handle != COGL_INVALID_HANDLE);
 
-  cogl_material_set_layer (ctx->default_material, 0, texture_handle);
+  cogl_material_set_layer (ctx->simple_material, 0, texture_handle);
   cogl_color_set_from_4ub (&white, 0xff, 0xff, 0xff, 0xff);
-  cogl_material_set_color (ctx->default_material, &white);
-  cogl_set_source (ctx->default_material);
+  cogl_material_set_color (ctx->simple_material, &white);
+  cogl_set_source (ctx->simple_material);
 }
 
 CoglMaterialFilter
@@ -1953,7 +2047,7 @@ cogl_material_layer_get_mag_filter (CoglHandle layer_handle)
 
 void
 cogl_material_set_layer_filters (CoglHandle         handle,
-                                 gint               layer_index,
+                                 int                layer_index,
                                  CoglMaterialFilter min_filter,
                                  CoglMaterialFilter mag_filter)
 {
