@@ -14,9 +14,9 @@
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * License along with this library. If not, see <http://www.gnu.org/licenses/>.
+ *
+ *
  */
 
 #ifdef HAVE_CONFIG_H
@@ -28,14 +28,10 @@
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
+#include <fcntl.h>
 
 #include <glib/gi18n-lib.h>
 
-#include <sys/stat.h>
-#include <sys/ioctl.h>
-#include <fcntl.h>
-#include <errno.h>
-
 #include <GL/glx.h>
 #include <GL/glxext.h>
 #include <GL/gl.h>
@@ -62,48 +58,6 @@ static ClutterBackendGLX *backend_singleton = NULL;
 
 static gchar    *clutter_vblank_name = NULL;
 
-#ifdef __linux__
-#define DRM_VBLANK_RELATIVE 0x1;
-
-struct drm_wait_vblank_request {
-    int           type;
-    unsigned int  sequence;
-    unsigned long signal;
-};
-
-struct drm_wait_vblank_reply {
-    int          type;
-    unsigned int sequence;
-    long         tval_sec;
-    long         tval_usec;
-};
-
-typedef union drm_wait_vblank {
-    struct drm_wait_vblank_request request;
-    struct drm_wait_vblank_reply reply;
-} drm_wait_vblank_t;
-
-#define DRM_IOCTL_BASE                  'd'
-#define DRM_IOWR(nr,type)               _IOWR(DRM_IOCTL_BASE,nr,type)
-#define DRM_IOCTL_WAIT_VBLANK           DRM_IOWR(0x3a, drm_wait_vblank_t)
-
-static int drm_wait_vblank(int fd, drm_wait_vblank_t *vbl)
-{
-    int ret, rc;
-
-    do
-      {
-        ret = ioctl(fd, DRM_IOCTL_WAIT_VBLANK, vbl);
-        vbl->request.type &= ~DRM_VBLANK_RELATIVE;
-        rc = errno;
-      }
-    while (ret && rc == EINTR);
-
-    return rc;
-}
-
-#endif
-
 G_CONST_RETURN gchar*
 clutter_backend_glx_get_vblank_method (void)
 {
@@ -262,6 +216,7 @@ clutter_backend_glx_get_features (ClutterBackend *backend)
   ClutterBackendGLX *backend_glx = CLUTTER_BACKEND_GLX (backend);
   const gchar *glx_extensions = NULL;
   ClutterFeatureFlags flags;
+  gboolean use_dri = FALSE;
 
   flags = clutter_backend_x11_get_features (backend);
   flags |= CLUTTER_FEATURE_STAGE_MULTIPLE;
@@ -275,7 +230,7 @@ clutter_backend_glx_get_features (ClutterBackend *backend)
                 "  GL_VENDOR: %s\n"
                 "  GL_RENDERER: %s\n"
                 "  GL_VERSION: %s\n"
-                "  GL_EXTENSIONS: %s\n",
+                "  GL_EXTENSIONS: %s",
                 glGetString (GL_VENDOR),
                 glGetString (GL_RENDERER),
                 glGetString (GL_VERSION),
@@ -283,109 +238,131 @@ clutter_backend_glx_get_features (ClutterBackend *backend)
 
   glx_extensions =
     glXQueryExtensionsString (clutter_x11_get_default_display (),
-			      clutter_x11_get_default_screen ());
+                              clutter_x11_get_default_screen ());
 
-  CLUTTER_NOTE (BACKEND, "GLX Extensions: %s", glx_extensions);
+  CLUTTER_NOTE (BACKEND, "  GLX Extensions: %s", glx_extensions);
+
+  use_dri = check_vblank_env ("dri");
 
   /* First check for explicit disabling or it set elsewhere (eg NVIDIA) */
-  if (getenv ("__GL_SYNC_TO_VBLANK") || check_vblank_env ("none"))
+  if (check_vblank_env ("none"))
     {
       CLUTTER_NOTE (BACKEND, "vblank sync: disabled at user request");
+      goto vblank_setup_done;
     }
-  else
+
+  if (g_getenv ("__GL_SYNC_TO_VBLANK") != NULL)
     {
-      /* We try two GL vblank syncing mechanisms.
-       * glXSwapIntervalSGI is tried first, then glXGetVideoSyncSGI.
-       *
-       * glXSwapIntervalSGI is known to work with Mesa and in particular
-       * the Intel drivers. glXGetVideoSyncSGI has serious problems with
-       * Intel drivers causing terrible frame rate so it only tried as a
-       * fallback.
-       *
-       * How well glXGetVideoSyncSGI works with other driver (ATI etc) needs
-       * to be investigated. glXGetVideoSyncSGI on ATI at least seems to have
-       * no effect.
-       */
-      if (!check_vblank_env ("dri") &&
-          _cogl_check_extension ("GLX_SGI_swap_control", glx_extensions))
+      backend_glx->vblank_type = CLUTTER_VBLANK_GLX_SWAP;
+      flags |= CLUTTER_FEATURE_SYNC_TO_VBLANK;
+
+      CLUTTER_NOTE (BACKEND, "Using __GL_SYNC_TO_VBLANK hint");
+      goto vblank_setup_done;
+    }
+
+  /* We try two GL vblank syncing mechanisms.
+   * glXSwapIntervalSGI is tried first, then glXGetVideoSyncSGI.
+   *
+   * glXSwapIntervalSGI is known to work with Mesa and in particular
+   * the Intel drivers. glXGetVideoSyncSGI has serious problems with
+   * Intel drivers causing terrible frame rate so it only tried as a
+   * fallback.
+   *
+   * How well glXGetVideoSyncSGI works with other driver (ATI etc) needs
+   * to be investigated. glXGetVideoSyncSGI on ATI at least seems to have
+   * no effect.
+   */
+  if (!use_dri &&
+      _cogl_check_extension ("GLX_SGI_swap_control", glx_extensions))
+    {
+      backend_glx->swap_interval =
+        (SwapIntervalProc) cogl_get_proc_address ("glXSwapIntervalSGI");
+
+      CLUTTER_NOTE (BACKEND, "attempting glXSwapIntervalSGI vblank setup");
+
+      if (backend_glx->swap_interval != NULL &&
+          backend_glx->swap_interval (1) == 0)
         {
-          backend_glx->swap_interval =
-            (SwapIntervalProc) cogl_get_proc_address ("glXSwapIntervalSGI");
+          backend_glx->vblank_type = CLUTTER_VBLANK_GLX_SWAP;
+          flags |= CLUTTER_FEATURE_SYNC_TO_VBLANK;
 
-          CLUTTER_NOTE (BACKEND, "attempting glXSwapIntervalSGI vblank setup");
-
-          if (backend_glx->swap_interval != NULL)
-            {
-              if (backend_glx->swap_interval (1) == 0)
-                {
-                  backend_glx->vblank_type = CLUTTER_VBLANK_GLX_SWAP;
-                  flags |= CLUTTER_FEATURE_SYNC_TO_VBLANK;
-                  CLUTTER_NOTE (BACKEND, "glXSwapIntervalSGI setup success");
-                }
-            }
-
-          if (!(flags & CLUTTER_FEATURE_SYNC_TO_VBLANK))
-            CLUTTER_NOTE (BACKEND, "glXSwapIntervalSGI vblank setup failed");
+          CLUTTER_NOTE (BACKEND, "glXSwapIntervalSGI setup success");
 
 #ifdef GLX_INTEL_swap_event
-          /* GLX_INTEL_swap_event allows us to avoid blocking the CPU while
-           * we wait for glXSwapBuffers to complete, and instead we get an X
-           * event notifying us of completion... */
-          if (!(clutter_debug_flags & CLUTTER_DEBUG_DISABLE_SWAP_EVENTS) &&
-              _cogl_check_extension ("GLX_INTEL_swap_event", glx_extensions) &&
-              flags & CLUTTER_FEATURE_SYNC_TO_VBLANK)
-            flags |= CLUTTER_FEATURE_SWAP_EVENTS;
-#endif /* GLX_INTEL_swap_event */
-        }
-
-      if (!check_vblank_env ("dri") &&
-          !(flags & CLUTTER_FEATURE_SYNC_TO_VBLANK) &&
-          _cogl_check_extension ("GLX_SGI_video_sync", glx_extensions))
-        {
-          CLUTTER_NOTE (BACKEND, "attempting glXGetVideoSyncSGI vblank setup");
-
-          backend_glx->get_video_sync =
-            (GetVideoSyncProc) cogl_get_proc_address ("glXGetVideoSyncSGI");
-
-          backend_glx->wait_video_sync =
-            (WaitVideoSyncProc) cogl_get_proc_address ("glXWaitVideoSyncSGI");
-
-          if ((backend_glx->get_video_sync != NULL) &&
-              (backend_glx->wait_video_sync != NULL))
+          /* GLX_INTEL_swap_event allows us to avoid blocking the CPU
+           * while we wait for glXSwapBuffers to complete, and instead
+           * we get an X event notifying us of completion...
+           */
+          if (!(clutter_paint_debug_flags & CLUTTER_DEBUG_DISABLE_SWAP_EVENTS) &&
+              _cogl_check_extension ("GLX_INTEL_swap_event", glx_extensions))
             {
-              CLUTTER_NOTE (BACKEND,
-                            "glXGetVideoSyncSGI vblank setup success");
-
-              backend_glx->vblank_type = CLUTTER_VBLANK_GLX;
-              flags |= CLUTTER_FEATURE_SYNC_TO_VBLANK;
+              flags |= CLUTTER_FEATURE_SWAP_EVENTS;
             }
+#endif /* GLX_INTEL_swap_event */
 
-          if (!(flags & CLUTTER_FEATURE_SYNC_TO_VBLANK))
-            CLUTTER_NOTE (BACKEND, "glXGetVideoSyncSGI vblank setup failed");
+          goto vblank_setup_done;
         }
 
+      CLUTTER_NOTE (BACKEND, "glXSwapIntervalSGI vblank setup failed");
+    }
+
+  if (!use_dri &&
+      !(flags & CLUTTER_FEATURE_SYNC_TO_VBLANK) &&
+      _cogl_check_extension ("GLX_SGI_video_sync", glx_extensions))
+    {
+      CLUTTER_NOTE (BACKEND, "attempting glXGetVideoSyncSGI vblank setup");
+
+      backend_glx->get_video_sync =
+        (GetVideoSyncProc) cogl_get_proc_address ("glXGetVideoSyncSGI");
+
+      backend_glx->wait_video_sync =
+        (WaitVideoSyncProc) cogl_get_proc_address ("glXWaitVideoSyncSGI");
+
+      if ((backend_glx->get_video_sync != NULL) &&
+          (backend_glx->wait_video_sync != NULL))
+        {
+          CLUTTER_NOTE (BACKEND, "glXGetVideoSyncSGI vblank setup success");
+
+          backend_glx->vblank_type = CLUTTER_VBLANK_GLX;
+          flags |= CLUTTER_FEATURE_SYNC_TO_VBLANK;
+
+          goto vblank_setup_done;
+        }
+
+      CLUTTER_NOTE (BACKEND, "glXGetVideoSyncSGI vblank setup failed");
+    }
+
 #ifdef __linux__
-      /*
-       * DRI is really an extreme fallback -rumoured to work with Via chipsets
-       */
-      if (!(flags & CLUTTER_FEATURE_SYNC_TO_VBLANK))
-        {
-          CLUTTER_NOTE (BACKEND, "attempting DRI vblank setup");
-          backend_glx->dri_fd = open("/dev/dri/card0", O_RDWR);
-          if (backend_glx->dri_fd >= 0)
-            {
-              CLUTTER_NOTE (BACKEND, "DRI vblank setup success");
-              backend_glx->vblank_type = CLUTTER_VBLANK_DRI;
-              flags |= CLUTTER_FEATURE_SYNC_TO_VBLANK;
-            }
+  /*
+   * DRI is really an extreme fallback -rumoured to work with Via chipsets
+   */
+  if (!(flags & CLUTTER_FEATURE_SYNC_TO_VBLANK))
+    {
+      CLUTTER_NOTE (BACKEND, "attempting DRI vblank setup");
 
-          if (!(flags & CLUTTER_FEATURE_SYNC_TO_VBLANK))
-            CLUTTER_NOTE (BACKEND, "DRI vblank setup failed");
+      backend_glx->dri_fd = open("/dev/dri/card0", O_RDWR);
+      if (backend_glx->dri_fd >= 0)
+        {
+          CLUTTER_NOTE (BACKEND, "DRI vblank setup success");
+
+          backend_glx->vblank_type = CLUTTER_VBLANK_DRI;
+          flags |= CLUTTER_FEATURE_SYNC_TO_VBLANK;
+
+          goto vblank_setup_done;
         }
+
+      CLUTTER_NOTE (BACKEND, "DRI vblank setup failed");
+    }
 #endif /* __linux__ */
 
-      if (!(flags & CLUTTER_FEATURE_SYNC_TO_VBLANK))
-        CLUTTER_NOTE (BACKEND, "no use-able vblank mechanism found");
+  CLUTTER_NOTE (BACKEND, "no use-able vblank mechanism found");
+
+vblank_setup_done:
+
+  if (_cogl_check_extension ("GLX_MESA_copy_sub_buffer", glx_extensions))
+    {
+      backend_glx->copy_sub_buffer =
+        (CopySubBufferProc) cogl_get_proc_address ("glXCopySubBufferMESA");
     }
 
   CLUTTER_NOTE (BACKEND, "backend features checked");
@@ -723,85 +700,16 @@ clutter_backend_glx_ensure_context (ClutterBackend *backend,
     }
 }
 
-static void
-glx_wait_for_vblank (ClutterBackendGLX *backend_glx)
-{
-  /* If we are going to wait for VBLANK manually, we not only need
-   * to flush out pending drawing to the GPU before we sleep, we
-   * need to wait for it to finish. Otherwise, we may end up with
-   * the situation:
-   *
-   *        - We finish drawing      - GPU drawing continues
-   *        - We go to sleep         - GPU drawing continues
-   * VBLANK - We call glXSwapBuffers - GPU drawing continues
-   *                                 - GPU drawing continues
-   *                                 - Swap buffers happens
-   *
-   * Producing a tear. Calling glFinish() first will cause us to properly
-   * wait for the next VBLANK before we swap. This obviously does not
-   * happen when we use GLX_SWAP and let the driver do the right thing
-   */
-
-  switch (backend_glx->vblank_type)
-    {
-    case CLUTTER_VBLANK_GLX_SWAP:
-      /* Nothing */
-      break;
-
-    case CLUTTER_VBLANK_GLX:
-      {
-        unsigned int retraceCount;
-
-        glFinish ();
-
-        backend_glx->get_video_sync (&retraceCount);
-        backend_glx->wait_video_sync (2,
-                                      (retraceCount + 1) % 2,
-                                      &retraceCount);
-      }
-      break;
-
-    case CLUTTER_VBLANK_DRI:
-#ifdef __linux__
-      {
-        drm_wait_vblank_t blank;
-
-        glFinish ();
-
-        blank.request.type     = DRM_VBLANK_RELATIVE;
-        blank.request.sequence = 1;
-        blank.request.signal   = 0;
-        drm_wait_vblank (backend_glx->dri_fd, &blank);
-      }
-#endif
-      break;
-
-    case CLUTTER_VBLANK_NONE:
-    default:
-      break;
-    }
-}
-
+/*
+ * FIXME: we should remove backend_class->redraw() and just
+ * have stage_window_iface->redraw()
+ */
 static void
 clutter_backend_glx_redraw (ClutterBackend *backend,
                             ClutterStage   *stage)
 {
-  ClutterBackendX11 *backend_x11;
-  ClutterStageGLX *stage_glx;
-  ClutterStageX11 *stage_x11;
-  ClutterStageWindow *impl;
-  CLUTTER_STATIC_TIMER (painting_timer,
-                        "Redrawing", /* parent */
-                        "Painting actors",
-                        "The time spent painting actors",
-                        0 /* no application private data */);
-  CLUTTER_STATIC_TIMER (swapbuffers_timer,
-                        "Redrawing", /* parent */
-                        "glXSwapBuffers",
-                        "The time spent blocked by glXSwapBuffers",
-                        0 /* no application private data */);
+  ClutterStageWindow *impl = _clutter_stage_get_window (stage);
 
-  impl = _clutter_stage_get_window (stage);
   if (G_UNLIKELY (impl == NULL))
     {
       CLUTTER_NOTE (BACKEND, "Stage [%p] has no implementation", stage);
@@ -810,40 +718,8 @@ clutter_backend_glx_redraw (ClutterBackend *backend,
 
   g_assert (CLUTTER_IS_STAGE_GLX (impl));
 
-  backend_x11 = CLUTTER_BACKEND_X11 (backend);
-  stage_x11 = CLUTTER_STAGE_X11 (impl);
-  stage_glx = CLUTTER_STAGE_GLX (impl);
-
-  CLUTTER_TIMER_START (_clutter_uprof_context, painting_timer);
-  /* this will cause the stage implementation to be painted */
-  clutter_actor_paint (CLUTTER_ACTOR (stage));
-  cogl_flush ();
-  CLUTTER_TIMER_STOP (_clutter_uprof_context, painting_timer);
-
-  if (stage_x11->xwin != None)
-    {
-      GLXDrawable drawable =
-        stage_glx->glxwin ? stage_glx->glxwin : stage_x11->xwin;
-
-      /* wait for the next vblank */
-      CLUTTER_NOTE (BACKEND, "Waiting for vblank");
-      glx_wait_for_vblank (CLUTTER_BACKEND_GLX (backend));
-
-      /* push on the screen */
-      CLUTTER_NOTE (BACKEND, "glXSwapBuffers (display: %p, window: 0x%lx)",
-                    backend_x11->xdpy,
-                    (unsigned long) drawable);
-
-      /* If we have GLX swap buffer events then glXSwapBuffers will return
-       * immediately and we need to track that there is a swap in
-       * progress... */
-      if (clutter_feature_available (CLUTTER_FEATURE_SWAP_EVENTS))
-        stage_glx->pending_swaps++;
-
-      CLUTTER_TIMER_START (_clutter_uprof_context, swapbuffers_timer);
-      glXSwapBuffers (backend_x11->xdpy, drawable);
-      CLUTTER_TIMER_STOP (_clutter_uprof_context, swapbuffers_timer);
-    }
+  clutter_stage_glx_redraw (CLUTTER_STAGE_GLX (impl),
+                            stage);
 }
 
 static ClutterStageWindow *
